/* Ocamlyacc parser for FFBB */

%{
open Ast
let lambda_num = ref 0
      
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE LSQUARE RSQUARE COMMA PLUS MINUS TIMES DIVIDE ASSIGN COLON IN
%token NOT EQ NEQ LT LEQ GT GEQ AND OR DOUBLEPLUS DOUBLEMINUS LAMBDA FUNC ARROW
%token RETURN IF ELSE FOR WHILE INT BOOL FLOAT VOID STRING LISTT DICTT SETT
%token <int> LITERAL
%token <bool> BLIT
%token <string> ID FLIT STRLIT 
%token EOF

%start program
%type <Ast.program> program

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN COLON
%left OR
%left AND
%left DOUBLEPLUS DOUBLEMINUS
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT
%left LSQUARE

%%

program:
  decls EOF { $1 }

decls:
   /* nothing */ { ([], [])               }
 | decls vdecl { (($2 :: fst $1), snd $1) }
 | decls fdecl { (fst $1, ($2 :: snd $1)) }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [($1,$2)]     }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

formals_opt_lambda:
    /* nothing */  { ([], []) }
  | formal_list_lambda   { (List.rev (fst $1), List.rev (snd $1)) }

formal_list_lambda:
    typ ID                   { ([$1],[$2])     }
  | formal_list_lambda COMMA typ ID { ($3 :: fst $1), ($4 :: snd $1) }

fdecl:
   typ ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
    { 
    { typ = $1;
	 fname = $2;
	 formals = List.rev $4;
	 body = List.rev $7 } 
   }


typ_list:
    typ                {[$1]}
  | typ COMMA typ_list { $1 :: $3 }

high_typ:
    LISTT { LIST }
  | SETT  { SET }
  | DICTT { DICT }

typ:
    INT      { Int   }
  | BOOL     { Bool  }
  | FLOAT    { Float }
  | STRING   { String }
  | VOID     { Void  }
  | FUNC LT typ_list GT { FunctionType($3) }
  | high_typ LT typ_list GT { CompositeType($1, $3) } 

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
    typ ID SEMI { ($1, $2) }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI                               { Expr $1                       }
  | RETURN expr_opt SEMI                    { Return $2                     }
  | LBRACE stmt_list RBRACE                 { Block(List.rev $2)            }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([]))         }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7)                }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
                                            { For($3, $5, $7, $9)           }
  | FOR ID IN expr stmt                       { Forin($2, $4, $5)             }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)                 }
  | typ ID SEMI                             { VarDecl($1, $2)               }
  | typ ID ASSIGN expr SEMI                 { VarDeclAssign(($1, $2), $4)   }
  | ID DOUBLEPLUS SEMI                      { Expr(Unop(Inc, Id($1)))       }
  | ID DOUBLEMINUS SEMI                     { Expr(Unop(Dec, Id($1)))       }
  | ID LSQUARE expr RSQUARE DOUBLEPLUS SEMI  { Expr(IndexUnop($1, $3, Inc)) }
  | ID LSQUARE expr RSQUARE DOUBLEMINUS SEMI { Expr(IndexUnop($1, $3, Dec)) }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    LITERAL          { Literal($1)            }
  | FLIT	         { Fliteral($1)           }
  | BLIT             { BoolLit($1)            }
  | STRLIT           { StrLit($1)          }
  | function_literal { $1 }
  | ID               { Id($1)                 }
  | expr PLUS   expr { Binop($1, Add,   $3)   }
  | expr MINUS  expr { Binop($1, Sub,   $3)   }
  | expr TIMES  expr { Binop($1, Mult,  $3)   }
  | expr DIVIDE expr { Binop($1, Div,   $3)   }
  | expr EQ     expr { Binop($1, Equal, $3)   }
  | expr NEQ    expr { Binop($1, Neq,   $3)   }
  | expr LT     expr { Binop($1, Less,  $3)   }
  | expr LEQ    expr { Binop($1, Leq,   $3)   }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3)   }
  | expr AND    expr { Binop($1, And,   $3)   }
  | expr OR     expr { Binop($1, Or,    $3)   }
  | list_expr        { $1 }
  | MINUS expr %prec NOT { Unop(Neg, $2)      }
  | NOT expr         { Unop(Not, $2)          }
  | ID LSQUARE expr RSQUARE ASSIGN expr { IndexAssign($1, $3, $6) }
  | ID ASSIGN expr   { Assign($1, $3)         }
  | ID LPAREN args_opt RPAREN { Call($1, $3)  }
  | ID LSQUARE expr COLON expr RSQUARE { GetSlice($1, $3, $5)  }
  | ID LSQUARE expr RSQUARE { GetIndex($1, $3)  }
  | LPAREN expr RPAREN { $2                   }

function_literal:
   typ LAMBDA formals_opt LBRACE stmt_list RBRACE{ 
        lambda_num := !lambda_num + 1;
        FunctionLit({
            typ = $1;
            fname = "lambda_" ^ string_of_int !lambda_num;
            formals = List.rev $3;
            body = List.rev $5
        }) 
    }
  | typ LAMBDA formals_opt ARROW expr { 
        lambda_num := !lambda_num + 1;
        FunctionLit({
            typ = $1;
            fname = "lambda_" ^ string_of_int !lambda_num;
            formals = List.rev $3;
            body = [Return $5]
        }) 
    }

list_expr:
    LSQUARE RSQUARE                { ListExpr([]) }
  | LSQUARE list_expr_core RSQUARE { ListExpr($2)}

list_expr_core:
    expr                      { [$1] }
  | expr COMMA list_expr_core { $1 :: $3 }

args_opt:
    /* nothing */ { [] }
  | args_list  { List.rev $1 }

args_list:
    expr                    { [$1] }
  | args_list COMMA expr { $3 :: $1 }
