(* Semantically-checked Abstract Syntax Tree and functions for printing it *)

open Ast

type sexpr = typ * sx

and sfunc_decl = {
    styp : typ;
    sfname : string;
    sftype : typ;
    sformals : bind list;
    slocals : bindtbl;
    sbody : sstmt list;
  }

and sx =
  | SLiteral of int
  | SFliteral of string
  | SBoolLit of bool
  | SStrLit of string
  | SFunctionLit of sfunc_decl 
  | SIndexAssign of string * sexpr * sexpr
  | SGetSlice of string * sexpr * sexpr
  | SGetIndex of string * sexpr
  | SId of string
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAssign of string * sexpr
  | SListExpr of sexpr list
  | SCall of string * sexpr list
  | SNoexpr

and sstmt =
  | SBlock of sstmt list
  | SExpr of sexpr
  | SReturn of sexpr
  | SIf of sexpr * sstmt * sstmt
  | SFor of sexpr * sexpr * sexpr * sstmt
  | SForin of sstmt * sstmt * sstmt * sexpr * sexpr * sexpr * sstmt
  | SWhile of sexpr * sstmt
  | SVarDecl of bind
  | SVarDeclAssign of bind * sexpr

and bindtbl = (string, typ) Hashtbl.t

and sprogram = bind list * sfunc_decl list

(* Pretty-printing functions *)

let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : "
  ^ (match e with
    | SLiteral l -> string_of_int l
    | SBoolLit true -> "true"
    | SBoolLit false -> "false"
    | SFliteral l -> l
    | SFunctionLit(l) -> string_of_typ l.styp ^ " " ^
      l.sfname ^ "(" ^ String.concat ", " (List.map string_of_formal l.sformals) ^
      ")\n{}\n" 
    | SStrLit(l) -> "\"" ^ l ^ "\""
    | SId s -> s
    | SIndexAssign (v, e1, e2) ->
        v ^ "[" ^ string_of_sexpr e1 ^ "] = " ^ string_of_sexpr e2
    | SGetIndex (e1, e2) -> e1 ^ "[" ^ string_of_sexpr e2 ^ "]"
    | SGetSlice(v, e1, e2) -> v ^ "[" ^ string_of_sexpr e1 ^ ":" ^ string_of_sexpr e2 ^ "]"
    | SBinop (e1, o, e2) ->
        string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
    | SUnop (o, e) -> string_of_uop o ^ string_of_sexpr e
    | SListExpr el ->
        "[ " ^ String.concat ", " (List.map string_of_sexpr el) ^ " ]"
    | SAssign (v, e) -> v ^ " = " ^ string_of_sexpr e
    | SCall (f, el) ->
        f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
    | SNoexpr -> "")
  ^ ")"

let rec string_of_sstmt = function
  | SBlock stmts ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr expr -> string_of_sexpr expr ^ ";\n"
  (* | SVarDecl(expr) -> "decalre " ^ string_of_sexpr expr ^ ";\n"; *)
  | SReturn expr -> "return " ^ string_of_sexpr expr ^ ";\n"
  | SIf (e, s, SBlock []) ->
      "if (" ^ string_of_sexpr e ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n" ^
      string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(e1, e2, e3, s) ->
      "for (" ^ string_of_sexpr e1  ^ " ; " ^ string_of_sexpr e2 ^ " ; " ^
      string_of_sexpr e3  ^ ") " ^ string_of_sstmt s
  | SForin(s1, s2, s3, e1, e2, e3, s) ->
      string_of_sstmt s1 ^ string_of_sstmt s2 ^ string_of_sstmt s3 ^ 
      "while(" ^ string_of_sexpr e1 ^ ") {\n" ^ string_of_sstmt s
      ^ ", " ^ string_of_sexpr e2 ^ ")" ^ string_of_sexpr e3 ^ "}\n"
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ string_of_sstmt s
  | SVarDecl(t, id) -> string_of_typ t ^ " " ^ id ^ ";\n" 
  | SVarDeclAssign((t, id), e) -> string_of_typ t ^ " " ^ id ^ " = " ^ string_of_sexpr e ^ ";\n"

let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map string_of_typ (List.map fst fdecl.sformals)) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars)
  ^ "\n"
  ^ String.concat "\n" (List.map string_of_sfdecl funcs)
