#include "treebasics.h"

struct TreeDict 
{
    int size;
    int key_type;  // 0: int, 1: bool, 2: double, 3: string
    int value_type;  // 0: int, 1: bool, 2: double, 3: string, 4: list
    struct TreeNode* root;
};

struct TreeDict* create_empty_dict(int key_type_id, int value_type_id)
{
    struct TreeDict* treedict = (struct TreeDict*)malloc(sizeof(struct TreeDict));
    treedict->size = 0;
    treedict->key_type = key_type_id;
    treedict->value_type = value_type_id;
    treedict->root = NULL;
    return treedict;
}

struct TreeNode* create_dict_node(struct TreeNode* parent, union Value key, union Value value)
{
    struct TreeNode* node = (struct TreeNode*)malloc(sizeof(struct TreeNode));
    node->key = key;
    node->value = value;
    node->parent = parent;
    node->lc = NULL;
    node->rc = NULL;
    return node;
}

void add_key_value(struct TreeDict* treedict, void* key_ptr, void* value_ptr) 
{
    union Value ukey = parse_value(key_ptr, treedict->key_type);
    union Value uvalue = parse_value(value_ptr, treedict->value_type);
    if (treedict->root == NULL)
    {
        treedict->root = create_dict_node(NULL, ukey, uvalue);
        treedict->size = 1;
        return;
    }
    struct TreeNode* curr_node = treedict->root;
    while (true)
    {
        if (uvalue_equal(ukey, curr_node->key, treedict->key_type)) 
        {
            curr_node->value = uvalue;
            return;
        }
        if (uvalue_lt(ukey, curr_node->key, treedict->key_type))
        {
            if (curr_node->lc == NULL)
            {
                curr_node->lc = create_dict_node(curr_node, ukey, uvalue);
                treedict->size++;
                return;
            }
            else
            {
                curr_node = curr_node->lc;
            }
        }
        else
        {
            if (curr_node->rc == NULL)
            {
                curr_node->rc = create_dict_node(curr_node, ukey, uvalue);
                treedict->size++;
                return;
            }
            else
            {
                curr_node = curr_node->rc;
            }
        }
    }
}

struct TreeNode* search_key_internal(struct TreeDict* treedict, union Value ukey) 
{
    if (treedict->root == NULL) return NULL;
    struct TreeNode* curr_node = treedict->root;
    while (true)
    {
        if (uvalue_equal(ukey, curr_node->key, treedict->key_type)) return curr_node;
        if (uvalue_lt(ukey, curr_node->key, treedict->key_type))
        {
            if (curr_node->lc == NULL) return NULL;
            else
            {
                curr_node = curr_node->lc;
            }
        }
        else
        {
            if (curr_node->rc == NULL) return NULL;
            else
            {
                curr_node = curr_node->rc;
            }
        }
    }
}

bool key_exists(struct TreeDict* treedict, void* key_ptr) 
{
    union Value ukey = parse_value(key_ptr, treedict->key_type);
    struct TreeNode* curr_node = search_key_internal(treedict, ukey);
    return (curr_node != NULL);
}

union Value retrieve_value_union(struct TreeDict* treedict, void* key_ptr) 
{
    union Value ukey = parse_value(key_ptr, treedict->key_type);
    struct TreeNode* curr_node = search_key_internal(treedict, ukey);
    if (curr_node == NULL)
    {
        printf("Error! Key %d does not exists\n.", ukey.i);
        exit(-1);
    }
    return curr_node->value;
}

int retrieve_value_int(struct TreeDict* treedict, void* key_ptr)
{
    return retrieve_value_union(treedict, key_ptr).i;
}

double retrieve_value_float(struct TreeDict* treedict, void* key_ptr)
{
    double result = retrieve_value_union(treedict, key_ptr).f;
    return result;
}

bool retrieve_value_bool(struct TreeDict* treedict, void* key_ptr)
{
    return retrieve_value_union(treedict, key_ptr).b;
}

struct LIST* retrieve_value_list(struct TreeDict* treedict, void* key_ptr)
{
    return retrieve_value_union(treedict, key_ptr).l;
}

void delete_key(struct TreeDict* treedict, void* key_ptr) 
{
    union Value ukey = parse_value(key_ptr, treedict->key_type);
    // reference: https://blog.csdn.net/zxnsirius/article/details/52131433
    struct TreeNode* curr_node = search_key_internal(treedict, ukey);
    if (curr_node == NULL) return;
    treedict->size--;
    if (curr_node->lc == NULL)
    {
        if (curr_node == treedict->root)  // root node
        {
            treedict->root = curr_node->rc;
            if (treedict->root != NULL) treedict->root->parent = NULL;
        }
        else if (curr_node->rc == NULL)  // leaf node
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = NULL;
            else curr_node->parent->rc = NULL;
        }
        else                             // middle node
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = curr_node->rc;
            else curr_node->parent->rc = curr_node->rc;
            curr_node->rc->parent = curr_node->parent;
        }
    }
    else if (curr_node->rc == NULL)
    {
        if (curr_node == treedict->root)
        {
            treedict->root = curr_node->lc;
            if (treedict->root != NULL) treedict->root->parent = NULL;
        }
        else
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = curr_node->lc;
            else curr_node->parent->rc = curr_node->lc;
            curr_node->lc->parent = curr_node->parent;
        }
    }
    else  // both left and right child exist, find the largest element that is smaller than the deleted element
    {
        struct TreeNode* left_neighbor = curr_node->lc;
        while (true)
        {
            if (left_neighbor->rc != NULL) left_neighbor = left_neighbor->rc;
            else break;
        }
        if (left_neighbor->lc == NULL)  // leaf node
        {
            if (is_left_child(left_neighbor)) left_neighbor->parent->lc = NULL;
            else left_neighbor->parent->rc = NULL;
        }
        else
        {
            if (is_left_child(left_neighbor)) left_neighbor->parent->lc = left_neighbor->lc;
            else left_neighbor->parent->rc = left_neighbor->lc;
            left_neighbor->lc->parent = left_neighbor->parent;
        }
        curr_node->key = left_neighbor->key;
        curr_node->value = left_neighbor->value;
    }
}

int get_dict_size(struct TreeDict* treedict)
{
    return treedict->size;
}


// int main(void)
// {
//     struct TreeDict* dict1 = create_empty_dict(0, 4);
//     int i3 = 3;
//     int i5 = 5;
//     struct LIST* list1 = create_empty_list(0);
//     struct LIST* list2 = range(5);
//     struct LIST* list3 = range(3);
//     add_key_value(dict1, (void*)(&i3), (void*)(&list1));
//     add_key_value(dict1, (void*)(&i5), (void*)(&list2));
//     add_key_value(dict1, (void*)(&i3), (void*)(&list3));
//     int dict1_size = get_dict_size(dict1);
//     printf("dict1 size: %d\n", dict1_size);
//     struct LIST* value1 = retrieve_value_list(dict1, (void*)(&i3));
//     printf("retrieved list size: %d\n", len((void*)(&value1)));
//     append((void*)(&value1), (void*)(&i5));
//     struct LIST* value2 = retrieve_value_list(dict1, (void*)(&i3));
//     printf("retrieved list size: %d\n", len((void*)(&value2)));
//     return 0;
// }
